from datetime import datetime
from typing import List, Dict

from pydantic import BaseModel

from edps.client.FitsFile import FitsFile
from edps.client.JobInfo import JobInfo
from edps.client.ProcessingJobStatus import ProcessingJobStatus


class LogEntry(BaseModel):
    file_name: str = None

    class Config:
        schema_extra = {
            "example": {
                "file_name": "esorex.log",
            }
        }

    @staticmethod
    def from_dict(d):
        return LogEntry(file_name=d['file_name'])


class ReportEntryPanel(BaseModel):
    file_name: str = None
    media_type: str = None
    description: str = None
    prodcatg: str = None

    class Config:
        schema_extra = {
            "example": {
                "file_name": "some_report.png",
                "media_type": "image/png",
                "description": "some report",
                "prodcatg": "ANCILLARY.PREVIEW"
            }
        }

    @staticmethod
    def from_dict(d) -> 'ReportEntryPanel':
        return ReportEntryPanel(file_name=d.get('file_name', None), media_type=d.get('media_type', None),
                                description=d.get('description', None), prodcatg=d.get('prodcatg', None))


class ReportEntry(BaseModel):
    report_name: str = None
    return_code: int = 0
    panels: List[ReportEntryPanel] = []

    class Config:
        schema_extra = {
            "example": {
                "report_name": "some_report",
                "return_code": 0,
                "panels": [ReportEntryPanel.Config.schema_extra['example']]
            }
        }

    @staticmethod
    def from_dict(d) -> 'ReportEntry':
        return ReportEntry(report_name=d.get('report_name', None), return_code=d.get('return_code', None),
                           panels=[ReportEntryPanel.from_dict(entry) for entry in d.get('panels', [])])


class ProcessingJob(BaseModel):
    completion_date: str = None
    submission_date: str = None
    status: ProcessingJobStatus = None
    input_files: List[FitsFile] = []
    output_files: List[FitsFile] = []
    logs: List[LogEntry] = []
    reports: List[ReportEntry] = []
    recipe_parameters: Dict[str, str] = {}
    configuration: JobInfo = None
    rejected: bool = False
    interrupted: bool = False

    @staticmethod
    def from_dict(d) -> 'ProcessingJob':
        return ProcessingJob(
            completion_date=d['completion_date'],
            submission_date=d['submission_date'],
            status=ProcessingJobStatus(d['status']),
            input_files=[FitsFile.from_dict(file) for file in d.get('input_files', [])],
            output_files=[FitsFile.from_dict(file) for file in d.get('output_files', [])],
            logs=[LogEntry.from_dict(log) for log in d.get('logs', [])],
            reports=[ReportEntry.from_dict(report) for report in d.get('reports', [])],
            recipe_parameters=d.get('recipe_parameters', {}),
            configuration=JobInfo.from_dict(d['configuration']),
            rejected=d.get('rejected', False),
            interrupted=d.get('interrupted', False),
        )

    def __repr__(self):
        return self.json()

    class Config:
        schema_extra = {
            "example": {
                "completion_date": datetime.now().isoformat(),
                "submission_date": datetime.now().isoformat(),
                "status": ProcessingJobStatus.COMPLETED,
                "input_files": [FitsFile.Config.schema_extra['example']],
                "output_files": [FitsFile.Config.schema_extra['example']],
                "reports": [ReportEntry.Config.schema_extra['example']],
                "logs": [LogEntry.Config.schema_extra['example']],
                "recipe_parameters": {"a": "b"},
                "configuration": JobInfo.Config.schema_extra['example'],
                "rejected": False,
                "interrupted": False
            }
        }
