import inspect
import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import List, Callable, Optional

from edps.client.WorkflowDTO import AssociationConfigurationDTO
from edps.utils import is_lambda
from .association_breakpoints import AssociationBreakpoints
from .fits import ClassifiedFitsFile
from .time_range import RelativeTimeRange

MatchFunction = Callable[[ClassifiedFitsFile, ClassifiedFitsFile], bool]
MatchKeywords = List[str]


def get_association_configuration_logger() -> Logger:
    return logging.getLogger('AssociationConfiguration')


@dataclass
class AssociationConfiguration:
    level: float
    match_function: MatchFunction
    match_keywords: MatchKeywords
    time_range: RelativeTimeRange
    logger: Logger = field(default_factory=get_association_configuration_logger)

    def get_text(self) -> str:
        return f"match({self.match_keywords})" if self.match_keywords else inspect.getsource(self.match_function)

    def get_match_function_name(self) -> Optional[str]:
        return self.match_function.__name__ if not is_lambda(self.match_function) else None

    def are_associated(self, ref: ClassifiedFitsFile, f: ClassifiedFitsFile,
                       breakpoints: AssociationBreakpoints, is_datasource: bool) -> bool:
        try:
            return (self.time_range.is_within_range(ref, f)
                    and self.match_function(ref, f)
                    and not breakpoints.is_association_crossing_breakpoint(ref, f, is_datasource))
        except Exception as e:
            self.logger.debug("association exception ref %s f %s : %s", ref.get_path(), f.get_path(), e, exc_info=e)
            return False

    def as_dto(self) -> AssociationConfigurationDTO:
        return AssociationConfigurationDTO(level=self.level, time_range=self.time_range.as_dto(),
                                           match_function=self.get_match_function_name(),
                                           match_keywords=self.match_keywords)
