import json
import logging
from bisect import bisect_left
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Dict

import requests

from edps.generator.constants import INSTRUME, PRO_CATG
from edps.generator.fits import ClassifiedFitsFile

logger = logging.getLogger("AssociationBreakpoint")

ANY = 'ANY'


@dataclass
class AssociationBreakpoint:
    instrument: str
    dpInstrument: str
    mjdObs: float
    proCatg: str
    rawType: str
    doClass: str

    @classmethod
    def from_dict(cls, d: Dict) -> 'AssociationBreakpoint':
        return cls(instrument=d['instrument'],
                   dpInstrument=d['dpInstrument'],
                   mjdObs=d['mjdObs'],
                   proCatg=d['proCatg'],
                   rawType=d['rawType'],
                   doClass=d['doClass'])


class AssociationBreakpoints:

    def __init__(self, breakpoints_in: List[Dict]):
        logger.debug("Parsing %d association breakpoints", len(breakpoints_in))
        breakpoints = [AssociationBreakpoint.from_dict(x) for x in breakpoints_in]
        self.breakpoints_by_ins_cat = defaultdict(lambda: defaultdict(list))
        for bp in breakpoints:
            self.breakpoints_by_ins_cat[bp.instrument][bp.doClass].append(bp.mjdObs)
            if bp.proCatg != bp.doClass:
                self.breakpoints_by_ins_cat[bp.instrument][bp.proCatg].append(bp.mjdObs)
        for ins in self.breakpoints_by_ins_cat:
            for cat in self.breakpoints_by_ins_cat[ins]:
                self.breakpoints_by_ins_cat[ins][cat].sort()

    def is_association_crossing_breakpoint(self, ref: ClassifiedFitsFile, f: ClassifiedFitsFile,
                                           is_datasource: bool) -> bool:
        if self.is_static_calib(is_datasource, f):
            return False
        min_mjd, max_mjd = sorted([ref.get_mjdobs(), f.get_mjdobs()])
        instrument = ref.get_keyword_value(INSTRUME, None)
        breakpoints_by_cat = self.breakpoints_by_ins_cat.get(instrument, {})
        breakpoints_for_cat = breakpoints_by_cat.get(f.classification, [])
        breakpoints_for_any = breakpoints_by_cat.get(ANY, [])
        return (self.find_breakpoint(min_mjd, max_mjd, breakpoints_for_cat) or
                self.find_breakpoint(min_mjd, max_mjd, breakpoints_for_any))

    @staticmethod
    def is_static_calib(is_datasource: bool, f: ClassifiedFitsFile) -> bool:
        return is_datasource and f.get_keyword_value(PRO_CATG, None) is not None

    @staticmethod
    def find_breakpoint(min_mjd: float, max_mjd: float, breakpoints: List[float]) -> bool:
        i = bisect_left(breakpoints, max_mjd)
        if i and breakpoints[i - 1] > min_mjd:
            return True
        return False

    @classmethod
    def from_url(cls, url: str) -> 'AssociationBreakpoints':
        logger.debug("Loading association breakpoints from URL %s", url)
        resp = requests.get(url, timeout=15)
        return cls(resp.json())

    @classmethod
    def from_file(cls, filename: str) -> 'AssociationBreakpoints':
        logger.debug("Loading association breakpoints from FILE %s", filename)
        with open(filename) as fid:
            return cls(json.load(fid))
