import os.path
from collections import defaultdict
from dataclasses import dataclass
from typing import Union, Dict, Set

import networkx as nx
from jinja2 import Environment, FileSystemLoader

from .meta_targets import QC1_CALIB, SCIENCE
from .task import Task, DataSource, AssociatedInput, CommandType

SIMPLE_GRAPH_TEMPLATE = "/templates/simple_graph.dot"
DETAILED_GRAPH_TEMPLATE = "/templates/detailed_graph.dot"
BLUE = 'deepskyblue3'
GREEN = 'green'
WHITE = 'white'


@dataclass(frozen=True)
class AssociatedInputInfo:
    name: str
    is_task: bool
    is_conditional: bool
    is_mandatory: bool

    @classmethod
    def from_associated_input(cls, inp: AssociatedInput):
        return cls(name=inp.input_task.name,
                   is_task=isinstance(inp.input_task, Task),
                   is_conditional=not inp.condition.__name__.startswith('<'),
                   is_mandatory=inp.min_ret > 0)


class Graph:
    def __init__(self, graph: nx.DiGraph, title: str):
        self.graph = graph
        self.tasks = [node for node in nx.topological_sort(self.graph) if isinstance(node, Task)]
        self.data_sources = [node for node in nx.topological_sort(self.graph) if isinstance(node, DataSource)]
        self.workflow_name = title
        self.tasks_by_subworkflow = defaultdict(list)
        self.meta_targets_by_subworkflow = defaultdict(set)
        for task in self.tasks:
            subworkflow = task.subworkflow_name[0] if task.subworkflow_name else "main"
            self.tasks_by_subworkflow[subworkflow].append(task)
            self.meta_targets_by_subworkflow[subworkflow].update(task.meta_targets)
        self.environment = Environment(loader=FileSystemLoader(os.path.dirname(__file__)),
                                       trim_blocks=True, lstrip_blocks=True)

    @staticmethod
    def get_name_or_subworkflow(node: Union[Task, DataSource]) -> str:
        if isinstance(node, Task) and node.subworkflow_name:
            return node.subworkflow_name[0]
        else:
            return node.name

    @property
    def task_children(self) -> Dict:
        task_edges = defaultdict(list)
        for source, target in self.graph.edges:
            source_name = self.get_name_or_subworkflow(source)
            target_name = self.get_name_or_subworkflow(target)
            if isinstance(source, Task) and source_name != target_name:
                task_edges[source_name].append(target_name)
        return dict(task_edges)

    @staticmethod
    def get_color_for_meta_targets(meta_targets: Set[str]) -> str:
        return BLUE if SCIENCE in meta_targets else GREEN if QC1_CALIB in meta_targets else WHITE

    def is_subworkflow(self, task: str) -> bool:
        return task in self.tasks_by_subworkflow

    def detailed_graph(self) -> str:
        template = self.environment.get_template(DETAILED_GRAPH_TEMPLATE)
        associated_tasks_by_task = {}
        associated_datasources_by_task = {}
        task_colors = {}
        task_commands = {}
        for task in self.tasks:
            associated_inputs = {AssociatedInputInfo.from_associated_input(inp) for inp in
                                 task.flatten_associated_inputs()}
            associated_tasks_by_task[task.name] = [inp for inp in associated_inputs if inp.is_task]
            associated_datasources_by_task[task.name] = [inp for inp in associated_inputs if not inp.is_task]
            task_colors[task.name] = self.get_color_for_meta_targets(task.meta_targets)
            if task.command_type == CommandType.FUNCTION:
                task_commands[task.name] = 'function.' + task.command.split('.')[-1]
            else:
                task_commands[task.name] = task.command

        subworkflow_colors = {subw: self.get_color_for_meta_targets(meta_targets) for subw, meta_targets in
                              self.meta_targets_by_subworkflow.items()}

        colors = [1, 11, 2, 10, 3, 9, 4, 8, 5, 7]
        edge_colors = {task: colors[i % len(colors)] for i, task in enumerate(self.task_children)}
        edge_colors_for_subworkflow_tasks = {}
        for parent, color in edge_colors.items():
            if self.is_subworkflow(parent):
                edge_colors_for_subworkflow_tasks.update(
                    {task.name: color for task in self.tasks_by_subworkflow[parent]})
        edge_colors.update(edge_colors_for_subworkflow_tasks)

        context = {
            "workflow": self.workflow_name,
            "tasks": self.tasks_by_subworkflow["main"],
            "task_colors": task_colors,
            "task_commands": task_commands,
            "subworkflow_colors": subworkflow_colors,
            "tasks_by_subworkflow": {k: v for k, v in self.tasks_by_subworkflow.items() if k != "main"},
            "associated_tasks_by_task": associated_tasks_by_task,
            "associated_datasources_by_task": associated_datasources_by_task,
            "task_children": self.task_children,
            "graph_edges": list(self.graph.edges),
            "edge_colors": edge_colors
        }
        return template.render(context)

    def simple_graph(self) -> str:
        template = self.environment.get_template(SIMPLE_GRAPH_TEMPLATE)
        raw_types = {ds.name for ds in self.data_sources if ds.is_main_input}
        tasks = {t.name for t in self.tasks if not t.subworkflow_name}
        subworkflows = {t.subworkflow_name[0] for t in self.tasks if t.subworkflow_name}
        edges = [(self.get_name_or_subworkflow(n0), self.get_name_or_subworkflow(n1))
                 for n0, n1 in self.graph.edges if n0 in self.tasks or n0 == n1.main_input]
        edges = [(e0, e1) for e0, e1 in edges if e0 != e1]
        colors = [1, 11, 2, 10, 3, 9, 4, 8, 5, 7]
        edge_colors = {}
        for i, (n0, n1) in enumerate(edges):
            if n0 in raw_types:
                edge_colors[n0] = colors[i % len(colors)]
        context = {
            "workflow": self.workflow_name,
            "raw_types": raw_types,
            "tasks": tasks,
            "subworkflows": subworkflows,
            "edges": edges,
            "edge_colors": edge_colors
        }
        return template.render(context)
