import hashlib
import os.path
import shutil
import tarfile
import logging
from typing import List, Tuple, Dict, Any, Optional, Set
from uuid import UUID

from astropy.io import fits
from astropy.io.fits import Header
from pydantic import BaseModel

from edps import FitsFile
from edps.client.ProcessingJob import ReportEntry
from edps.executor.renamer import PatternProductRenamer, ProductRenamer
from edps.generator.fits import FitsFileFactory
from edps.interfaces.JobsRepository import JobDetails, JobsRepository

AdditionalKeywords = Dict[str, Any]


class Phase3ScienceFileConfiguration(BaseModel):
    associated_fits_categories: Optional[List[str]] = []
    associated_report_categories: Optional[List[str]] = []
    tar_fits_categories: Optional[List[str]] = []
    tar_report_categories: Optional[List[str]] = []
    tar_category: Optional[str] = None


class Phase3StreamConfiguration(BaseModel):
    tasks: List[str]
    science_files_configuration: Dict[str, Phase3ScienceFileConfiguration]


class Phase3Configuration(BaseModel):
    output_location: str
    jobs_with_keywords: Dict[str, AdditionalKeywords]
    stream_configuration: Dict[str, Phase3StreamConfiguration]
    renaming_pattern: str


logger = logging.getLogger("build_phase3_dataset")


# DFS-21738
def build_phase3_dataset(global_configuration: Phase3Configuration, repository: JobsRepository) -> List[str]:
    result = []
    for stream_name, stream_config in global_configuration.stream_configuration.items():
        output_dir = os.path.join(global_configuration.output_location, stream_name)
        output = build_phase3_stream(output_dir, global_configuration.jobs_with_keywords, stream_config, global_configuration.renaming_pattern, repository)
        result.append(output)
    return result


# DFS-21368
def build_phase3_stream(output_location: str,
                        jobs_with_keywords: Dict[str, AdditionalKeywords],
                        stream_configuration: Phase3StreamConfiguration,
                        renaming_pattern: str,
                        repository: JobsRepository) -> str:
    job_ids = [UUID(job_id) for job_id in jobs_with_keywords.keys()]
    verify_leaves(job_ids, repository)
    renamer = PatternProductRenamer(renaming_pattern)
    os.makedirs(output_location)
    jobs = [repository.get_job_details(job_id) for job_id in job_ids]
    for job in jobs:
        if job.configuration.task_name in stream_configuration.tasks:
            logger.info("Processing job: %s %s", job.configuration.job_id, job.configuration.task_name)
            additional_keywords = jobs_with_keywords[job.configuration.job_id]
            science_files = [file.name for file in job.result.output_files if is_science_file(file.name)]
            associated_files = get_associated_files(job, repository)
            for main_file in science_files:
                logger.info("Processing science file: %s", main_file)
                renamed_main_file = renamer.rename(main_file)
                renamed_main_prefix, _ = os.path.splitext(os.path.basename(renamed_main_file))
                main_file_prodcatg = get_prodcatg(main_file)
                configuration = stream_configuration.science_files_configuration.get(main_file_prodcatg)
                if configuration:
                    with fits.open(main_file) as hdu_list:
                        job_dir = os.path.dirname(main_file)
                        primary_header = hdu_list[0].header
                        next_index = 1

                        # add keywords to primary header
                        for key, value in additional_keywords.items():
                            primary_header[key] = value

                        # add assoc files, remember to change prodcatg if it was SCIENCE
                        adjusted_assoc_files = []
                        for assoc_file in associated_files:
                            logger.info("Processing assoc file: %s", assoc_file)
                            assoc_file_prodcatg = get_prodcatg(assoc_file.name)
                            if assoc_file_prodcatg in configuration.associated_fits_categories:
                                logger.info("Assoc file: %s is configured to be added", assoc_file)
                                target_file = renamer.rename_file(assoc_file)
                                set_keyword(main_file, primary_header, "ASSON", next_index, target_file.name)
                                assoc_file_destination = os.path.join(output_location, target_file.name)
                                shutil.copy(assoc_file.name, assoc_file_destination)
                                logger.info("Assoc file: %s copied to %s", assoc_file, assoc_file_destination)
                                if assoc_file_prodcatg.startswith("SCIENCE"):
                                    logger.info("Assoc file: %s is science so the category will change to ANCILLARY", assoc_file)
                                    with fits.open(assoc_file_destination, mode='update') as target_assoc_file:
                                        assoc_primary_header = target_assoc_file[0].header
                                        assoc_primary_header['PRODCATG'] = assoc_file_prodcatg.replace("SCIENCE", "ANCILLARY")
                                        remove_asso_keywords(assoc_primary_header, 1)
                                adjusted_assoc_files.append(FitsFile(name=assoc_file_destination, category=assoc_file_prodcatg))
                                next_index += 1
                            else:
                                adjusted_assoc_files.append(assoc_file)

                        # add assoc reports
                        report_index = 1
                        for report in job.result.reports:
                            logger.info("Processing report: %s", report.report_name)
                            for panel in report.panels:
                                if panel.prodcatg in configuration.associated_report_categories:
                                    logger.info("Adding report panel: %s to main file", panel)
                                    panel_path = os.path.join(job_dir, report.report_name, panel.file_name)
                                    panel_name = rename_panel(renamed_main_prefix, panel, report_index)
                                    set_keyword(main_file, primary_header, "ASSON", next_index, panel_name)
                                    set_keyword(main_file, primary_header, "ASSOC", next_index, panel.prodcatg)
                                    set_keyword(main_file, primary_header, "ASSOM", next_index, file_md5(panel_path))
                                    panel_output = os.path.join(output_location, panel_name)
                                    shutil.copy(panel_path, panel_output)
                                    next_index += 1
                                    report_index += 1

                        # add tar, but the associated files in tar should already have the prodcatg adjusted so checksums match
                        if configuration.tar_category is not None:
                            tar_result = build_tar(main_file, adjusted_assoc_files, job.result.reports,
                                                   output_location, configuration, renamer)
                            if tar_result:
                                tar_name, tar_checksum = tar_result
                                set_keyword(main_file, primary_header, "ASSON", next_index, tar_name)
                                set_keyword(main_file, primary_header, "ASSOC", next_index, configuration.tar_category)
                                set_keyword(main_file, primary_header, "ASSOM", next_index, tar_checksum)
                                next_index += 1

                        remove_asso_keywords(primary_header, next_index)

                        hdu_list.writeto(os.path.join(output_location, renamed_main_file), checksum=True)
    return output_location


def rename_panel(main_prefix, panel, report_index):
    _, ext = os.path.splitext(panel.file_name)
    if report_index == 1:
        return f"{main_prefix}{ext}"
    else:
        return f"{main_prefix}_{report_index}{ext}"


def set_keyword(main_file: str, header: Header, keyword: str, index: int, value: str):
    if index > 999:
        raise RuntimeError(f"Exceeded index 999 for {keyword}={value} in {main_file}")
    header[f"{keyword}{index}"] = value


def build_tar(main_file: str, associated_files: List[FitsFile], reports: List[ReportEntry],
              output_location: str, configuration: Phase3ScienceFileConfiguration, renamer: ProductRenamer) -> Optional[Tuple[str, str]]:
    renamed_main_file = renamer.rename(main_file)
    renamed_main_prefix, _ = os.path.splitext(os.path.basename(renamed_main_file))
    tar_name = f"{renamed_main_prefix}.tar"
    tar_path = os.path.join(output_location, tar_name)
    logger.info("Building tar: %s", tar_path)
    tar = tarfile.open(tar_path, "w")
    job_dir = os.path.dirname(main_file)
    non_empty = False
    for file in associated_files:
        logger.info("Processing assoc file: %s for tar", file)
        if any([keyword in configuration.tar_fits_categories for keyword in get_tar_match_keywords(file.name)]):
            logger.info("Adding assoc file: %s to tar", file)
            tar.add(file.name, renamer.rename(file.name))
            non_empty = True
    report_index = 1
    for report in reports:
        logger.info("Processing report: %s for tar", report.report_name)
        for panel in report.panels:
            if panel.prodcatg in configuration.tar_report_categories:
                logger.info("Adding report panel: %s to tar", panel)
                panel_name = rename_panel(renamed_main_prefix, panel, report_index)
                tar.add(os.path.join(job_dir, report.report_name, panel.file_name), panel_name)
                non_empty = True
                report_index += 1
    tar.close()
    if non_empty:
        return tar_name, file_md5(tar_path)
    else:
        os.remove(tar_path)
        return None


def file_md5(path) -> str:
    with open(path, "rb") as f:
        return hashlib.md5(f.read()).hexdigest()


def get_associated_files(job: JobDetails, repository: JobsRepository, include_science=False) -> Set[FitsFile]:
    if job.configuration.input_files:
        non_main_outputs = [file for file in job.result.output_files if (include_science or not is_science_file(file.name))]
        # FIXME: we only take associated files and not outputs of associated jobs here, so it might not work for all cases?
        return set(non_main_outputs + job.configuration.associated_files)
    else:
        input_jobs = [repository.get_job_details(UUID(job_id)) for job_id in job.configuration.input_job_ids]
        recursive_associated_files = [list(get_associated_files(parent_job, repository, include_science=True)) for parent_job in input_jobs]
        return set(sum(recursive_associated_files, []))


def verify_leaves(job_ids: List[UUID], repository: JobsRepository):
    jobs = [repository.get_job_details(job_id) for job_id in job_ids]
    job_ids = set([str(job_id) for job_id in job_ids])
    for job in jobs:
        if job.configuration.input_job_ids:
            intersection = set(job.configuration.input_job_ids).intersection(job_ids)
            if intersection:
                raise RuntimeError(f"Jobs {intersection} are not leaves, they are main parents of {job.configuration.job_id}")


def is_science_file(path: str) -> bool:
    prodcatg_value = get_prodcatg(path)
    return prodcatg_value and prodcatg_value.startswith("SCIENCE")


def get_prodcatg(path: str) -> Optional[str]:
    prodcatg_keyword = "PRODCATG"
    keywords = FitsFileFactory.extract_provided_keywords(path, {prodcatg_keyword})
    return keywords.get(prodcatg_keyword, None)


def get_tar_match_keywords(path: str) -> Set[str]:
    return set(FitsFileFactory.extract_provided_keywords(path, {"PRODCATG", "PRO.CATG"}).values())


def remove_asso_keywords(primary_header: Header, index: int):
    while True:
        try:
            primary_header.remove("ASSOC" + str(index), ignore_missing=True)
            primary_header.remove("ASSOM" + str(index), ignore_missing=True)
            primary_header.remove("ASSON" + str(index))
            index += 1
        except KeyError:
            break
